/* Send a file to a socket */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>

#define DEFAULT_HOST "localhost"
#define DEFAULT_PORT 1999
#define BUFFER_SIZE 8192

#define STDIN_NAME "-"

/* Print a usage message */
static void usage(int argc, char *argv[])
{
    fprintf(stderr, "usage: %s [OPTION]... [FILE]...\n", argv[0]);
    fprintf(stderr, "  -h host\n");
    fprintf(stderr, "  -p port\n");
    fprintf(stderr, "  -t\n");
    fprintf(stderr, "  -u\n");
    fprintf(stderr, "  -v\n");
}


/* Copies bytes from one file desciptor to another */
static void do_copy(int in_fd, int out_fd)
{
    char buffer[BUFFER_SIZE];
    char *pointer;
    char *end;
    ssize_t length;

    while (1)
    {
	/* Read the next chunk of bytes */
	if ((length = read(in_fd, buffer, BUFFER_SIZE)) < 0)
	{
	    perror("read(): failed");
	    exit(1);
	}

	/* Watch for end of file */
	if (length == 0)
	{
	    return;
	}

	/* Write the bytes */
	pointer = buffer;
	end = pointer + length;
	while (pointer < end)
	{
	    if ((length = write(out_fd, pointer, end - pointer)) < 0)
	    {
		perror("write(): failed");
		exit(1);
	    }

	    pointer += length;
	}
    }
}

/* Reads characters from stdin and sends them to a socket */
int main(int argc, char *argv[])
{
    char *hostname = DEFAULT_HOST;
    int port = DEFAULT_PORT;
    int type = SOCK_STREAM;
    int choice, sock;
    struct sockaddr_in name;

    /* Parse the command line */
    while ((choice = getopt(argc, argv, "h:p:tuv")) > 0)
    {
	switch (choice)
	{
	    /* host */
	    case 'h':
	    {
		hostname = optarg;
		break;
	    }

	    /* port */
	    case 'p':
	    {
		/* FIX THIS: should look for a service with the given name */
		if ((port = atoi(optarg)) == 0)
		{
		    printf("%s: bad port number\n", optarg);
		    exit(1);
		}

		break;
	    }

	    /* TCP */
	    case 't':
	    {
		type = SOCK_STREAM;
		break;
	    }

	    /* UDP */
	    case 'u':
	    {
		type = SOCK_DGRAM;
		break;
	    }

	    /* version */
	    case 'v':
	    {
		printf("%s version %s\n", PACKAGE, VERSION);
		exit(0);
	    }

	    /* Bogus option --print usage */
	    default:
	    {
		usage(argc, argv);
		exit(1);
	    }
	}
    }

    /* Construct the IP address */
    memset(&name, 0, sizeof(name));
    name.sin_family = AF_INET;
    name.sin_port = htons(port);

    /* If a hostname was provided then look it up */
    if (hostname != NULL)
    {
	/* See if it's in dot notation */
	if ((name.sin_addr.s_addr = inet_addr(hostname)) == (unsigned long)-1)
	{
	    struct hostent *host;

	    /* Check with the name service */
	    while ((host = gethostbyname(hostname)) == NULL)
	    {
		switch (h_errno)
		{
		    case HOST_NOT_FOUND:
		    {
			fprintf(stderr, "%s: unknown host\n", hostname);
			exit(1);
		    }

		    case TRY_AGAIN:
		    {
			break;
		    }

		    case NO_RECOVERY:
		    case NO_DATA:
		    {
			fprintf(stderr, "%s: internal error\n", hostname);
			exit(1);
		    }
		}
	    }

	    name.sin_addr = *(struct in_addr *)host -> h_addr;
	}
    }

    /* Create a socket */
    if ((sock = socket(AF_INET, type, 0)) < 0)
    {
	perror("unable to create socket");
	exit(1);
    }

    /* Connect it */
    if (connect(sock, (struct sockaddr *)&name, sizeof(name)) < 0)
    {
	perror("unable to connect");
	exit(1);
    }

    /* If it's a multicast socket then join the group */
    if (IN_MULTICAST(htonl(name.sin_addr.s_addr)))
    {
	struct ip_mreq request;

	request.imr_multiaddr.s_addr = name.sin_addr.s_addr;
	request.imr_interface.s_addr = INADDR_ANY;
	if (setsockopt(
		sock,
		IPPROTO_IP, IP_ADD_MEMBERSHIP,
		&request, sizeof(request)) < 0)
	{
	    perror("setsockopt() failed");
	}
    }

    /* If no files were specified then pretend that stdin was */
    if (! (optind < argc))
    {
	argv[--optind] = STDIN_NAME;
    }

    /* Read from files (or stdin) and write to the socket */
    while (optind < argc)
    {
	char *filename;
	int fd;

	/* Open the next file */
	filename = argv[optind++];

	/* Are we reading from stdin */
	if (strcmp(filename, STDIN_NAME) == 0)
	{
	    fd = STDIN_FILENO;
	}
	else
	{
	    if ((fd = open(filename, O_RDONLY)) < 0)
	    {
		fprintf(stderr, "unable to open file %s: %s\n", filename, strerror(errno));
		exit(1);
	    }
	}

	/* Copy the file to the socket */
	do_copy(fd, sock);

	/* Close off the file and move on to the next one */
	if (close(fd) < 0)
	{
	    perror("close(): failed");
	    exit(1);
	}
    }

    /* Close the socket and exit */
    if (close(sock) < 0)
    {
	perror("close(): failed");
	exit(1);
    }

    exit(0);
}
