#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <netdb.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#define BUFFER_SIZE 8192
#define MAX_PENDING 30
#define DEFAULT_PORT 1999
#define PREFIX_LENGTH 16

int fd_map[FD_SETSIZE];

/* Prints the buffer in both hex and ASCII */
static void hexdump(char *prefix, char *buffer, ssize_t length)
{
    unsigned char *start = (unsigned char *)buffer;
    unsigned char *end = start + length;
    unsigned char *pointer;

    for (pointer = start - ((int)buffer & 0xF); pointer < end; pointer += 16)
    {
	int i;

	fprintf(stdout, "%s%p ", prefix, pointer);

	/* Print the next 16 bytes in hex */
	for (i = 0; i < 16; i++)
	{
	    if (i % 8 == 0)
	    {
		putchar(' ');
	    }

	    if (start <= pointer + i && pointer + i < end)
	    {
		fprintf(stdout, "%02x ", *(pointer + i));
	    }
	    else
	    {
		fprintf(stdout, "   ");
	    }
	}

	/* Print the same 16 bytes as ASCII */
	putchar('>');
	for (i = 0; i < 16; i++)
	{
	    if (start <= pointer + i && pointer + i < end)
	    {
		int ch = *(pointer + i);
		if (ch < 32 || ch > 126)
		{
		    putchar('.');
		}
		else
		{
		    putchar(ch);
		}
	    }
	    else
	    {
		putchar(' ');
	    }
	}

	fprintf(stdout, "<\n");
    }
}

/* Print a usage message */
static void usage(int argc, char *argv[])
{
    fprintf(stderr, "usage: %s [OPTIONS]...\n", argv[0]);
    fprintf(stderr, "  Source options:\n");
    fprintf(stderr, "   -h host\n");
    fprintf(stderr, "   -p port\n");
    fprintf(stderr, "\n");

    fprintf(stderr, "  Destination options:\n");
    fprintf(stderr, "   -H host\n");
    fprintf(stderr, "   -P port\n");
    fprintf(stderr, "\n");

    fprintf(stderr, "  Traffic class options:\n");
    fprintf(stderr, "   -t Listen for TCP traffic. The default.\n");
    fprintf(stderr, "   -u Listen for UDP traffic.\n");
    fprintf(stderr, "\n");

    fprintf(stderr, "  Output options:\n");
    fprintf(stderr, "   -x Dump each packet to stdout in hexadecimal format. The default.\n");
    fprintf(stderr, "   -r Dump each packet to stdout as raw ASCII.\n");
    fprintf(stderr, "   -q Do not print packet data.\n");
    fprintf(stderr, "   -qq Only print error conditions\n");
    fprintf(stderr, "\n");
    
    fprintf(stderr, "  Miscellaneous:\n");
    fprintf(stderr, "   -v Print version and quit.\n");
}

int main(int argc, char *argv[])
{
    char *hostname = NULL;
    int port = DEFAULT_PORT;
    char *fwd_hostname = NULL;
    int fwd_port = DEFAULT_PORT;
    int type = SOCK_STREAM;
    char *tname;
    int do_hex = 1;
    int quiet = 0;
    struct sockaddr_in name;
    struct sockaddr_in fwd_name;
    int choice, value, fd, i;
    fd_set read_fds;

    /* Ignore signals */
    if (signal(SIGPIPE, SIG_IGN) < 0) {
	fprintf(stdout, "signal(): failed: %s\n", strerror(errno));
	exit(1);
    }

    /* Parse the command-line */
    while ((choice = getopt(argc, argv, "h:p:H:P:rqtuxv")) > 0)
    {
	switch (choice)
	{
	    /* host */
	    case 'h':
	    {
		hostname = optarg;
		break;
	    }

	    /* port */
	    case 'p':
	    {
		/* FIX THIS: should look for a service with the given name */
		if ((port = atoi(optarg)) == 0)
		{
		    fprintf(stdout, "%s: bad port number\n", optarg);
		    exit(1);
		}

		break;
	    }

	    /* forwarding host */
	    case 'H':
	    {
		fwd_hostname = optarg;
		break;
	    }

	    /* forwarding port */
	    case 'P':
	    {
		/* FIX THIS: should look for a service with the given name */
		if ((fwd_port = atoi(optarg)) == 0)
		{
		    fprintf(stdout, "%s: bad port number\n", optarg);
		    exit(1);
		}

		break;
	    }

	    /* Quiet */
	    case 'q':
	    {
		quiet++;
		break;
	    }

	    /* TCP */
	    case 't':
	    {
		type = SOCK_STREAM;
		break;
	    }

	    /* UDP */
	    case 'u':
	    {
		type = SOCK_DGRAM;
		break;
	    }

	    /* raw */
	    case 'r':
	    {
		do_hex = 0;
		break;
	    }

	    /* hex */
	    case 'x':
	    {
		do_hex = 1;
		break;
	    }

	    /* version */
	    case 'v':
	    {
		fprintf(stdout, "%s version %s\n", PACKAGE, VERSION);
		exit(0);
	    }

	    /* errors... */
	    default:
	    {
		usage(argc, argv);
		exit(1);
	    }
	}
    }

    /* Construct the host address */
    memset(&name, 0, sizeof(struct sockaddr_in));
    name.sin_family = AF_INET;
    name.sin_port = htons(port);

    /* If a hostname was provided then look it up */
    if (hostname != NULL)
    {
	/* See if it's in dot notation */
	if ((name.sin_addr.s_addr = inet_addr(hostname)) == (unsigned long)-1)
	{
	    struct hostent *host;

	    /* Check with the name service */
	    while ((host = gethostbyname(hostname)) == NULL)
	    {
		switch (h_errno)
		{
		    case HOST_NOT_FOUND:
		    {
			fprintf(stdout, "%s: unknown host\n", hostname);
			exit(1);
		    }

		    case TRY_AGAIN:
		    {
			break;
		    }

		    default:
		    {
			fprintf(stdout, "%s: internal error\n", hostname);
			exit(1);
		    }
		}
	    }
	    name.sin_addr = *(struct in_addr *)host->h_addr;
	}
    }


    /* If we're forwarding packets then look up the target */
    if (fwd_hostname != NULL)
    {
	/* Construct the forwarding address */
	memset(&fwd_name, 0, sizeof(struct sockaddr_in));
	fwd_name.sin_family = AF_INET;
	fwd_name.sin_port = htons(fwd_port);

	/* Is the hostname in dot notation? */
	if ((fwd_name.sin_addr.s_addr = inet_addr(fwd_hostname)) == (unsigned long)-1)
	{
	    struct hostent *host;

	    /* Check with the name service */
	    while ((host = gethostbyname(fwd_hostname)) == NULL)
	    {
		switch (h_errno)
		{
		    case HOST_NOT_FOUND:
		    {
			fprintf(stdout, "%s: unknown host\n", fwd_hostname);
			exit(1);
		    }

		    case TRY_AGAIN:
		    {
			break;
		    }

		    default:
		    {
			fprintf(stdout, "%s: internal error\n", fwd_hostname);
			exit(1);
		    }
		}
	    }

	    fwd_name.sin_addr = *(struct in_addr *)host->h_addr;
	}
    }

    
    /* Create a socket */
    if ((fd = socket(AF_INET, type, 0)) < 0)
    {
	fprintf(stdout, "socket(): failed: %s\n", strerror(errno));
	exit(1);
    }

    /* Set the socket address to be reusable */
    value = 1;
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &value, sizeof(value)) < 0)
    {
	fprintf(stdout, "setsockopt(): failed: %s\n", strerror(errno));
    }

    /* Bind to the address */
    if (bind(fd, (struct sockaddr *)&name, sizeof(name)) < 0)
    {
	fprintf(stdout, "bind(): failed: %s\n", strerror(errno));
	exit(1);
    }

    if (type == SOCK_STREAM)
    {
	/* Listen for incoming connections */
	if (listen(fd, MAX_PENDING) < 0)
	{
	    fprintf(stdout, "listen(): failed: %s\n", strerror(errno));
	    exit(1);
	}

	tname = "TCP connections";
    }
    /* If we're in the multicast address range then join the multicast group */
    else if (IN_MULTICAST(htonl(name.sin_addr.s_addr)))
    {
	struct ip_mreq request;

	request.imr_multiaddr.s_addr = name.sin_addr.s_addr;
	request.imr_interface.s_addr = INADDR_ANY;
	if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &request, sizeof(request)) < 0)
	{
	    fprintf(stdout, "setsockopt(): failed: %s\n", strerror(errno));
	}

	tname = "multicast packets";
    }
    else
    {
	tname = "UDP packets";
    }

    if (quiet < 2)
    {
        fprintf(stdout, "--- Awaiting %s on %s:%d\n",
		tname, inet_ntoa(name.sin_addr), port);
    }

    /* Initialize the fd_map */
    for (i = 0; i < FD_SETSIZE; i++)
    {
	fd_map[i] = -1;
    }
    
    /* Initialize the read set */
    FD_ZERO(&read_fds);
    FD_SET(fd, &read_fds);


    /* Loop forever */
    while (1)
    {
	fd_set pending_fds;

	/* Figure out what's got input for us */
	memcpy(&pending_fds, &read_fds, sizeof(pending_fds));
	while (select(FD_SETSIZE - 1, &pending_fds, NULL, NULL, NULL) < 0)
	{
	    if (errno != EAGAIN)
	    {
		fprintf(stdout, "select(): failed: %s\n", strerror(errno));
		exit(1);
	    }

	    fprintf(stdout, "select(): failed: %s\n", strerror(errno));
	}

	/* Print out any pending input */
	for (i = 0; i < FD_SETSIZE; i++)
	{
	    /* Is there input on file descriptor i? */
	    if (FD_ISSET(i, &pending_fds))
	    {
		/* Watch for new TCP connections */
		if (type == SOCK_STREAM && i == fd)
		{
		    int size = sizeof(name);
		    int sock;

		    /* Accept the connection */
		    if ((sock = accept(fd, (struct sockaddr *)&name, &size)) < 0)
		    {
			fprintf(stdout, "accept(): failed: %s\n", strerror(errno));
			abort();
		    }

		    /* Add it to the read fds */
		    FD_SET(sock, &read_fds);

		    /* Print out new connection info */
                    if (quiet < 2) {
                        fprintf(stdout, "--- new connection %d: %s\n", sock, inet_ntoa(name.sin_addr));
                    }

		    /* If we're forwarding packets then make a new connection */
		    if (fwd_hostname != NULL)
		    {
			int s;

			/* Create a socket */
			if ((s = socket(AF_INET, type, 0)) < 0)
			{
			    fprintf(stdout, "unable to create socket: %s\n", strerror(errno));
			    FD_CLR(sock, &read_fds);
			    close(sock);
			}
			else if (connect(s, (struct sockaddr *)&fwd_name, sizeof(fwd_name)) < 0)
			{
			    fprintf(stdout, "unable to connect to forwarding host: %s\n", strerror(errno));
			    close(s);
			    FD_CLR(sock, &read_fds);
			    close(sock);
			}
			else
			{
			    /* Set up the map between sockets */
			    FD_SET(s, &read_fds);
			    fd_map[s] = sock;
			    fd_map[sock] = s;
			}
		    }
		}
		else
		{
		    char buffer[BUFFER_SIZE];
		    ssize_t length;
		    int sock = fd_map[i];

		    /* Read from the socket */
		    if (type == SOCK_DGRAM)
		    {
			struct sockaddr_in from;
			int fromlen = sizeof(from);

			if ((length = recvfrom(
			    i, buffer, BUFFER_SIZE, 0,
			    (struct sockaddr *)&from, &fromlen)) < 0)
			{
			    fprintf(stdout, "recvfrom(): failed: %s\n", strerror(errno));
			    exit(1);
			}

                        if (quiet < 2) {
                            fprintf(stdout, "--- read %d bytes from %s\n",
                                    length, inet_ntoa(from.sin_addr));
                        }
		    }
		    else
		    {
			/* Read from the socket */
			if ((length = read(i, buffer, BUFFER_SIZE)) < 0)
			{
			    fprintf(stdout, "read(): failed: %s\n", strerror(errno));
			    length = 0;
			}

			/* Check for zero-length read */
			if (length == 0)
			{
			    FD_CLR(i, &read_fds);

                            if (quiet < 2) {
                                fprintf(stdout, "--- %d closed\n", i);
                            }
			    close(i);

			    /* Close the forwarding connection */
			    if (! (sock < 0))
			    {
				FD_CLR(sock, &read_fds);
				close(sock);
				fd_map[sock] = -1;
				fd_map[i] = -1;
			    }
			}
			else
			{
                            if (quiet < 2) {
                                fprintf(stdout, "--- read %d bytes\n", length);
                            }
			}
		    }

		    if (length > 0)
		    {
			if (! quiet)
			{
			    if (do_hex)
			    {
				char prefix[PREFIX_LENGTH];

				/* Construct prefix for the hex dump */
				sprintf(prefix, "%d:", i);
				hexdump(prefix, buffer, length);
			    }
			    else
			    {
				if (fwrite(buffer, 1, length, stdout) == 0)
				{
				    fprintf(stdout, "fwrite(): failed: %s, ", strerror(errno));
				    exit(1);
				}
				fflush(stdout);
			    }
			}

			/* Forward the bytes */
			if (! (sock < 0))
			{
			    char *end = buffer + length;
			    char *pointer = buffer;

			    while (pointer < end)
			    {
				if ((length = write(sock, pointer, end - pointer)) < 0)
				{
				    fprintf(stdout, "write(): failed: %s\n", strerror(errno));
				    exit(1);
				}

				pointer += length;
			    }
			}
		    }
		}
	    }
	}
    }
}
